"""

Temporal values and comparison, including the attributes:
NEVER: a value that compares unequal to any value, including itself.
DISTANT_PAST: a value that compares less than any other value but itself and NEVER
FAR_FUTURE: a value that compares greater than any other value but itself and NEVER
"""

from __future__ import annotations

import typing

import numpy
import numpy.typing

__all__: list[str] = ["DISTANT_PAST", "FAR_FUTURE", "NEVER", "isnever"]

@typing.overload
def isnever(t: typing.SupportsFloat) -> bool:
    """
    Returns whether the value of t corresponds to "never". As "never" is implemented as a floating-point NaN,
    direct comparison will always fail, since NaN != NaN.
    """

@typing.overload
def isnever(t: typing.Annotated[numpy.typing.ArrayLike, numpy.float64]) -> numpy.typing.NDArray[numpy.bool]:
    """
    Returns an array of booleans corresponding to whether the element of an array correspond to "never". As "never" is
    implemented as a floating-point NaN, direct comparison will always fails, since NaN != NaN.
    """

DISTANT_PAST: float  # value = -inf
FAR_FUTURE: float  # value = inf
NEVER: float  # value = nan



import importlib.metadata

__version__ = importlib.metadata.version("neworder")

from _neworder_core import (
    CalendarTimeline,
    LinearTimeline,
    Model,
    MonteCarlo,
    NoTimeline,
    NumericTimeline,
    Timeline,
    checked,
    df,
    freethreaded,
    log,
    mpi,
    run,
    stats,
    thread_id,
    time,
    verbose,
)

# type: ignore
from .domain import Domain, Edge, Space, StateGrid
from .mc import as_np

__all__ = ["as_np", "Domain", "Space", "freethreaded", "thread_id"]



"""

Submodule for basic MPI environment discovery, containing the following attributes:

RANK: the process rank (0 in serial mode)
SIZE: the number of processes (1 in serial mode)
COMM: the MPI communicator (None in serial mode)
"""

from __future__ import annotations

import mpi4py.MPI  # type: ignore[import-not-found]

__all__: list[str] = ["COMM", "RANK", "SIZE"]
COMM: mpi4py.MPI.Intracomm  # value = <mpi4py.MPI.Intracomm object>
RANK: int = 0
SIZE: int = 1



import numpy as np

from neworder import MonteCarlo


def as_np(mc: MonteCarlo) -> np.random.Generator:
    """
    Returns an adapter enabling the MonteCarlo object to be used with numpy random functionality
    """

    class _NpAdapter(np.random.BitGenerator):
        def __init__(self, rng: MonteCarlo):
            super().__init__(0)
            self.rng = rng
            self.rng.init_bitgen(self.capsule)

    return np.random.Generator(_NpAdapter(mc))



"""

Submodule for operations involving direct manipulation of pandas dataframes
"""

from __future__ import annotations

import typing

import numpy
import numpy.typing

import neworder

__all__: list[str] = ["testfunc", "transition", "unique_index"]

def testfunc(model: neworder.Model, df: typing.Any, colname: str) -> None:
    """
    Test function for direct dataframe manipulation. Results may vary. Do not use.
    """

def transition(
    model: neworder.Model,
    categories: typing.Annotated[numpy.typing.ArrayLike, numpy.int64],
    transition_matrix: typing.Annotated[numpy.typing.ArrayLike, numpy.float64],
    df: typing.Any,
    colname: str,
) -> None:
    """
    Randomly changes categorical data in a dataframe, according to supplied transition probabilities.
    Args:
        model: The model (for access to the MonteCarlo engine).
        categories: The set of possible categories
        transition_matrix: The probabilities of transitions between categories
        df: The dataframe, which is modified in-place
        colname: The name of the column in the dataframe
    """

def unique_index(n: typing.SupportsInt, offset: typing.SupportsInt = 0) -> numpy.typing.NDArray[numpy.int64]:
    """
    Generates an array of n unique values, even across multiple processes, that can be used to unambiguously index multiple dataframes.
    When multiple threads are in use, a unique offset must be specifically provided for each thread. (Using the thread
    id is generally nondeterministic)
    """
